import importlib.util
import glob
import os.path
import sys
import traceback

import finishline.grid_components as gc

from dash.dependencies import Input, Output
import dash_core_components as dcc
import dash_html_components as html
import dash_building_blocks as dbb
import random
import json
from dash.exceptions import PreventUpdate



class FinishLine(object):
    
    def __init__(
            self, 
            app=None,
            data=None,
            name='default',
            debug=False,
            on_layout_change=None,
            debug_path=None):

        self.name = name
        
        # server side
        self.app = app or dash.Dash()
        self.data = data or {}
        self.plugins = {}
        self.blocks = BlockManager()
        self.store = FinishStore(app, hide=(not debug))
        
        # client side
        self.client_vis  = {}
        self.client_data = {}
        
        # misc
        self.extra_files = []
        self.debug = debug
        self.debug_path = debug_path
        
        #private
        self._curr_file = None
        
        
        
        # callbacks
        self.on_layout_change = on_layout_change or (lambda lo: print('layout', lo)) 
        
        
    def register_vis(self, name, layout):
                
        self.client_vis[name] = {
            'layout': layout,
            'src_file': self._curr_file
        }
        

    def register_data(self, name, inputs=None, state=None, data=None, on_update=None):
        
        self.client_data[name] = {
            'data': data or {},
            'src_file': self._curr_file
        }
                
        ret = self.store.register(name, inputs=inputs, state=state, initially=data)
        
        if on_update:
            @self.app.callback(
                Output(self.store.ids[name], 'role'),
                [self.store.input(name)]
            )
            def data_callback(new_data):
                on_update(json.loads(new_data))
                raise PreventUpdate()
        
        return ret
                
        
    def generate_layout(self, components=gc, layouts=None, cols=None):
        
        page_id = self.name + '-fl-page'
        
        page_layout = page_id + '-layout'
        page_config = page_id + '-config'
        page_data   = page_id + '-data'
        
        @self.register_data(
            page_config, 
            inputs=[Input(page_layout, 'layouts')],
            data=layouts,
            on_update=self.on_layout_change
        )
        def get_layout(new_config):
            return json.dumps(new_config)
                
        # client side data objects
        c_data_style = {'display':'block'} if self.debug else {'display':'none'}
        
        if self.debug_path is not None:
            c_data = self.store.debug_layout(self.client_data)
        else:
            c_data = self.store.layout
                
        # client side visualization objects
        c_vis = self._gen_c_vis(components, layouts)
        layout = components.Page(
            [components.Layout(c_vis, id=page_layout, layouts=layouts, cols=cols), 
             html.Div(c_data, className='fl-data', id=page_data, style=c_data_style)],
            id=page_id)

        # run plugin finalize method, e.g. should be used to create callbacks
        self._finalize()
        
        return layout
    
    
    def _gen_c_vis(self, components, layouts):
        i = 0
        c = []
        ii = [l['i'] for l in layouts['lg']] if (layouts and 'lg' in layouts) else []
        for (name, vis) in self.client_vis.items():
            key = str(i) if str(i) in ii else name
            key = name if name in ii else key
            c.append(components.Card(vis['layout'], title=name, i=key, href=vis['src_file']))
            i = i + 1
            
        return c
                
    
    def load_plugins(self, plugins_path='plugins/*'):
        
        modules = sorted(glob.glob(plugins_path))

        for m in modules:
#             print(m)
            fname = m + '/__init__.py'
            if not os.path.isfile(fname):
                continue
            
            if self.debug_path is not None:
                self._curr_file = os.path.abspath(fname).replace(self.debug_path['root'], self.debug_path['target'])
            
            self.extra_files.append(fname) #TODO walk all py files in dir
            spec = importlib.util.spec_from_file_location(m, fname)
#             print(spec)
            plugin = importlib.util.module_from_spec(spec)
            
            try:
                spec.loader.exec_module(plugin)
                plugin.initialize(self.app, self.data, self)
            except:
                traceback.print_exc()
                print("Unexpected error in plugin, ", m, ": ", sys.exc_info()[0])
                self.register_vis(m, html.Pre("Unexpected error in " + m + '\n' + traceback.format_exc()))
                # TODO: register 'XXX' instead of 'plugin/XXX'
                
            self.plugins[m] = plugin
             
                
    def _finalize(self):
        for plugin in self.plugins.values():
            if 'finalize' in plugin.__dict__:
                plugin.finalize(self.app, self.data, self)
    
    
    def run_server(self,
                   port=5000,
                   debug=False,
                   **flask_run_options):
        self.app.run_server(port=port, debug=debug, extra_files=self.extra_files, **flask_run_options)


class FinishStore(dbb.Store):
    
    def debug_layout(self, client_data):
        style = {'display': 'none'} if self.hide else None
        return html.Div([
            html.A(children=[k+':', html.Div(json.dumps(v['data']), 
                                                  id=self.ids[k])], 
                   href=v['src_file'], 
                   target=k) 
            for k, v in client_data.items()
        ])
        
    
class BlockManager:
    
    def __init__(self):
        
        self._blocks = {}
        
        
    def register(self, name, block):
        
        self._blocks[name] = block
        
        
    def __getitem__(self, name):
        
        return self._blocks[name]
        
        
    def __getattr__(self, name):
        
        return self._blocks[name]



# -*- coding: utf-8 -*-
import dash_core_components as dcc
import dash_html_components as html
from textwrap import dedent


def merge(a, b):
    return dict(a, **b)


def omit(omitted_keys, d):
    return {k: v for k, v in d.items() if k not in omitted_keys}


def Page(children=None, **kwargs):
    return html.Div(
        children,
        className='container-width',
        **kwargs
    )


def Layout(children=None, layouts=None, cols=None, rowHeight=300, **kwargs):    
    return html.Div(
        children,
        className="row",
        style={
            'padding': 10,
        },
        **kwargs
    )


def Card(children, title='Undefined', i=0, href=None, width=6, **kwargs):

    t = html.A(title, href=href, target=title) if href is not None else title
    c = [html.Div(t, className='fl-titlebar'), html.Div(children, style={'padding': 20 })]

    return html.Div(
        c,
        className='fl-card col-sm-{}'.format(str(width)),
        style=merge({
            'padding': 20,
            'margin': 0,
        }, kwargs.get('style', {})),
        **omit(['style'], kwargs)
    )


# -*- coding: utf-8 -*-
import dash_core_components as dcc
import dash_html_components as html
from textwrap import dedent

import dash_responsive_grid_layout as dgl
import copy

def merge(a, b):
    return dict(a, **b)


def omit(omitted_keys, d):
    return {k: v for k, v in d.items() if k not in omitted_keys}


def Page(children=None, **kwargs):
    return html.Div(
        children,
        className='fl-page',
        **kwargs
    )


def Layout(children=None, layouts=None, cols=None, rowHeight=300, **kwargs):\
    
    layouts = layouts or {'lg':[], 'md':[], 'sm':[] }
    cols = cols or { 'lg': 3, 'md': 2, 'sm': 1 };
    
    return dgl.ResponsiveGridLayout(
        children,
        layouts=layouts,
        cols=cols,
        rowHeight=rowHeight,
        draggableHandle='.fl-titlebar',
        **kwargs
    )


def Card(children, title='Undefined', i=0, href=None, **kwargs):
    
    # note: don't put children in a div container wrapper, else plotly won't resize properly
    #c = [html.Div(title, className='fl-titlebar'), html.Div(children, className='fl-content')]
    t = html.A(title, href=href, target=title) if href is not None else title
    c = [html.Div(t, className='fl-titlebar')] + [children]

    return html.Div(
        c,
        className='fl-card',
        key=str(i),
        style=merge({}, kwargs.get('style', {})),
        **omit(['style'], kwargs)
    )


from setuptools import setup

def readme():
    with open('README.rst') as readme_file:
        return readme_file.read()

configuration = {
    'name' : 'finishline',
    'version': '0.0.3',
    'description' : 'Framework for Building Beautiful and Functional Dashbords',
    'long_description' : readme(),
    'classifiers' : [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved',
        'Programming Language :: Python',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 3 :: Only',
    ],
    'keywords' : 'dash dashboard ui grid layout',
    'url' : 'http://github.com/AlgorithmHub/finishline',
    'maintainer' : 'Alex Cabello',
    'maintainer_email' : 'alex.cabello@algorithmhub.com',
    'license' : 'MIT',
    'packages' : ['finishline'],
    'install_requires': ['dash >= 0.22.0',
                         'dash-responsive-grid-layout >= 0.0.1',
                         'dash-building-blocks >= 0.0.2'],
    'ext_modules' : [],
    'cmdclass' : {},
    'test_suite' : '',
    'tests_require' : [],
    'data_files' : ()
    }

setup(**configuration)
